from __future__ import annotations

from typing import Any

from fastapi import FastAPI
from tortoise import Tortoise

from .settings import Settings


def build_tortoise_config(settings: Settings) -> dict[str, Any]:
    if settings.orm.apps:
        return {
            "connections": settings.orm.connections,
            "apps": settings.orm.apps,
        }

    # default app config - users can extend via settings
    return {
        "connections": settings.orm.connections,
        "apps": {
            "models": {
                "models": settings.orm.models,
                "default_connection": "default",
            }
        },
    }


def init_db(app: FastAPI, settings: Settings) -> None:
    config = build_tortoise_config(settings)

    @app.on_event("startup")
    async def init_orm() -> None:
        await Tortoise.init(config=config)

    @app.on_event("shutdown")
    async def close_orm() -> None:
        await Tortoise.close_connections()



from __future__ import annotations

from tortoise import fields
from tortoise.models import Model


class Example(Model):
    id = fields.IntField(pk=True)
    name = fields.CharField(max_length=100)



from __future__ import annotations

import os
from pathlib import Path

from typer.testing import CliRunner

from fast_django.cli.main import app

runner = CliRunner()


def test_runserver_infer_module(tmp_path: Path) -> None:
    proj = tmp_path / "mysite"
    (proj).mkdir()
    (proj / "asgi.py").write_text("app=None\n", encoding="utf-8")
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        os.environ["FD_TEST_MODE"] = "1"
        res = runner.invoke(app, ["runserver", "--port", "9999", "--host", "127.0.0.1"])  # type: ignore[list-item]
        # Since uvicorn won't start here, command should still return 0 (we don't check)
        assert res.exit_code == 0



from __future__ import annotations

import anyio

from fast_django import Settings, create_app
from fast_django.orm import Tortoise
from fast_django.scaffolding.project.project_name.models import User
from fast_django.signals import post_delete, post_save, pre_delete, pre_save


async def _exercise_orm() -> None:
    # Ensure app bootstraps ORM
    s = Settings()
    s.installed_apps = ["fast_django.scaffolding.project.project_name"]
    s.orm.models = [
        "fast_django.scaffolding.project.project_name.models",
        "aerich.models",
    ]
    app = create_app(s)
    # Trigger startup to init ORM and create tables
    await app.router.startup()
    await Tortoise.generate_schemas()

    events: list[str] = []

    def on_pre_save(sender, **kwargs):
        if kwargs.get("created"):
            events.append("pre_create")
        else:
            events.append("pre_update")

    def on_post_save(sender, **kwargs):
        if kwargs.get("created"):
            events.append("post_create")
        else:
            events.append("post_update")

    def on_pre_delete(sender, **kwargs):
        events.append("pre_delete")

    def on_post_delete(sender, **kwargs):
        events.append("post_delete")

    pre_save.connect(on_pre_save)
    post_save.connect(on_post_save)
    pre_delete.connect(on_pre_delete)
    post_delete.connect(on_post_delete)

    # Create
    u = User(email="a@example.com")
    await u.save()
    assert "pre_create" in events and "post_create" in events

    events.clear()
    # Update
    u.is_active = False
    await u.save()
    assert "pre_update" in events and "post_update" in events

    events.clear()
    # Delete
    await u.delete()
    assert "pre_delete" in events and "post_delete" in events

    # Cleanly shutdown
    await app.router.shutdown()


def test_orm_signals() -> None:
    anyio.run(_exercise_orm)



from __future__ import annotations

import asyncio
import importlib
import os
import sys
import types
from pathlib import Path
from typing import Any

from fastapi import FastAPI
from typer.testing import CliRunner

from fast_django.admin import try_call_admin_hooks
from fast_django.apps import include_app_routers
from fast_django.cli.main import app
from fast_django.settings import Settings

runner = CliRunner()


def test_startproject_scaffold(tmp_path: Path) -> None:
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        res = runner.invoke(app, ["startproject", "mysite"])  # type: ignore[list-item]
        assert res.exit_code == 0
        assert (Path("mysite") / "orm_config.py").exists()
        assert (Path("mysite") / "mysite" / "asgi.py").exists()
        assert (Path("mysite") / "mysite" / "models.py").exists()


def test_startapp_scaffold(tmp_path: Path) -> None:
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        res = runner.invoke(app, ["startapp", "blog"])  # type: ignore[list-item]
        assert res.exit_code == 0
        assert (Path("blog") / "models.py").exists()


def test_runserver_reload_false(tmp_path: Path) -> None:
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        os.environ["FD_TEST_MODE"] = "1"
        (Path("mysite")).mkdir()
        (Path("mysite") / "asgi.py").write_text("app=None\n", encoding="utf-8")
        res = runner.invoke(app, ["runserver", "--no-reload"])  # type: ignore[list-item]
        assert res.exit_code == 0


def test_try_call_admin_hooks_swallow_exceptions(monkeypatch: Any) -> None:
    called: dict[str, bool] = {"x": False}
    mod = types.ModuleType("boom.admin")

    def _hook(*args: Any, **kwargs: Any) -> None:
        called["x"] = True
        raise RuntimeError("boom")

    mod.init_admin = _hook  # type: ignore[attr-defined]
    monkeypatch.setitem(__import__("sys").modules, "boom.admin", mod)
    settings = Settings()
    settings.installed_apps = ["boom"]
    app_fast = FastAPI()
    try_call_admin_hooks(app_fast, settings)
    assert called["x"] is True


def test_include_app_routers_missing_module() -> None:
    app_fast = FastAPI()
    include_app_routers(app_fast, ["missing.module.name"])
    # should not raise and routes remain empty
    assert app_fast.routes == app_fast.routes


def test_makemigrations_skip_init(tmp_path: Path, monkeypatch: Any) -> None:
    calls: list[list[str]] = []

    def fake_run(cmd: list[str], check: bool = False) -> None:  # type: ignore[no-untyped-def]
        calls.append(cmd)

    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        Path("pyproject.toml").write_text("[project]\nname='x'\n", encoding="utf-8")
        Path("aerich.ini").write_text("[aerich]\n", encoding="utf-8")
        monkeypatch.setattr("subprocess.run", fake_run)
        res = runner.invoke(app, ["makemigrations"])  # type: ignore[list-item]
        assert res.exit_code == 0
        assert any(cmd[:2] == ["aerich", "migrate"] for cmd in calls)
        assert not any(cmd[:2] == ["aerich", "init"] for cmd in calls)


def test_createsuperuser_success(tmp_path: Path, monkeypatch: Any) -> None:
    # create a fake models module with Tortoise-like API
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        (Path("myproj")).mkdir()
        (Path("myproj") / "__init__.py").write_text("\n", encoding="utf-8")
        (Path("myproj") / "models.py").write_text(
            """
from fast_django.cli.main import DoesNotExist

class User:
    created = []
    @classmethod
    async def get(cls, email: str):
        raise DoesNotExist
    @classmethod
    async def create(cls, **kwargs):
        cls.created.append(kwargs)
            """.strip(),
            encoding="utf-8",
        )

        # monkeypatch Tortoise and CryptContext and run_async
        class DummyTortoise:
            @staticmethod
            async def init(config: dict[str, Any]) -> None:
                return None

            @staticmethod
            async def close_connections() -> None:
                return None

        class DummyCrypt:
            def __init__(self, *args: Any, **kwargs: Any) -> None:
                pass

            def hash(self, pw: str) -> str:
                return "hashed-" + pw

        def run_sync(coro):  # type: ignore[no-untyped-def]
            return asyncio.run(coro)

        class DummyDoesNotExist(Exception):
            pass

        monkeypatch.setattr("tortoise.Tortoise", DummyTortoise)
        monkeypatch.setattr("tortoise.run_async", run_sync)
        monkeypatch.setattr("fast_django.cli.main.CryptContext", DummyCrypt)
        monkeypatch.setattr("fast_django.cli.main.DoesNotExist", DummyDoesNotExist)

        sys.path.insert(0, str(Path.cwd()))
        res = runner.invoke(
            app,
            [
                "createsuperuser",
                "--email",
                "a@example.com",
                "--password",
                "pass",
                "--models",
                "myproj.models",
            ],
        )  # type: ignore[list-item]
        assert res.exit_code in (0, 1)
        # ensure our fake model recorded creation
        mod = importlib.import_module("myproj.models")
        assert mod.User.created and mod.User.created[0]["email"] == "a@example.com"


def test_createsuperuser_user_exists(tmp_path: Path, monkeypatch: Any) -> None:
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        (Path("myproj")).mkdir()
        (Path("myproj") / "__init__.py").write_text("\n", encoding="utf-8")
        (Path("myproj") / "models.py").write_text(
            """
class User:
    @classmethod
    async def get(cls, email: str):
        return object()
            """.strip(),
            encoding="utf-8",
        )

        class DummyTortoise:
            @staticmethod
            async def init(config: dict[str, Any]) -> None:
                return None

            @staticmethod
            async def close_connections() -> None:
                return None

        def run_sync(coro):  # type: ignore[no-untyped-def]
            return asyncio.run(coro)

        monkeypatch.setattr("tortoise.Tortoise", DummyTortoise)
        monkeypatch.setattr("tortoise.run_async", run_sync)
        # No need to patch CryptContext as we won't hit create branch
        sys.path.insert(0, str(Path.cwd()))
        res = runner.invoke(
            app,
            [
                "createsuperuser",
                "--email",
                "a@example.com",
                "--password",
                "pass",
                "--models",
                "myproj.models",
            ],
        )  # type: ignore[list-item]
        assert res.exit_code == 0



from __future__ import annotations

import shutil
from pathlib import Path

from typer.testing import CliRunner

from fast_django.cli.main import app

runner = CliRunner()


def test_startproject_tmp(tmp_path: Path) -> None:
    # Note: keep tmp dir isolated; no need to capture cwd
    try:
        # run inside tmp
        os_cwd = tmp_path
        (os_cwd).mkdir(parents=True, exist_ok=True)
        with runner.isolated_filesystem(temp_dir=str(os_cwd)):
            result = runner.invoke(app, ["startproject", "mysite"])  # type: ignore[list-item]
            assert result.exit_code == 0, result.output
            assert (Path("mysite") / "manage.py").exists()
            assert (Path("mysite") / "mysite" / "asgi.py").exists()
    finally:
        shutil.rmtree(tmp_path, ignore_errors=True)



from __future__ import annotations

from pydantic import Field

from fast_django.settings import OrmConfig, Settings as BaseSettings


class Settings(BaseSettings):
    app_name: str = "project_name"
    debug: bool = True
    orm: OrmConfig = OrmConfig(models=["project_name.models", "aerich.models"])
    installed_apps: list[str] = Field(default_factory=lambda: ["project_name"])



from __future__ import annotations

import anyio
import httpx

from fast_django import Settings, create_app


async def _request_healthz() -> None:
    s = Settings()
    s.installed_apps = ["fast_django.scaffolding.project.project_name"]  # use template module
    app = create_app(s)
    transport = httpx.ASGITransport(app=app)
    async with httpx.AsyncClient(transport=transport, base_url="http://test") as client:
        resp = await client.get("/healthz")
        assert resp.status_code == httpx.codes.OK
        assert resp.json()["status"] == "ok"


def test_app_healthz_route() -> None:
    anyio.run(_request_healthz)



from __future__ import annotations

from pathlib import Path

from typer.testing import CliRunner

from fast_django.cli.main import app as cli
from fast_django.orm import Model, fields
from fast_django.routers import APIRouter, FastAPI


def test_reexports() -> None:
    class Example(Model):
        id = fields.IntField(pk=True)

    router = APIRouter()
    api = FastAPI()
    assert router and api


def test_startapp_scaffold_files(tmp_path: Path) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=str(tmp_path)):
        res = runner.invoke(cli, ["startapp", "shop"])  # type: ignore[list-item]
        assert res.exit_code == 0
        assert (Path("shop") / "models.py").exists()
        assert (Path("shop") / "views.py").exists()
        assert (Path("shop") / "routes.py").exists()
        assert (Path("shop") / "migrations" / "__init__.py").exists()



from __future__ import annotations

from fastapi import FastAPI

from fast_django.admin import AdminSite
from fast_django.settings import Settings


def init_admin(app: FastAPI, settings: Settings) -> None:
    site = AdminSite(title="Admin")
    site.mount(app, settings)



from __future__ import annotations

from contextlib import suppress

import anyio
import httpx

from fast_django import Settings, create_app
from fast_django.signals import SignalsMiddleware, got_request_exception, request_finished


def test_non_http_passthrough() -> None:
    called = {"ok": False}

    async def asgi_app(scope, receive, send):  # type: ignore[no-redef]
        called["ok"] = True

    mw = SignalsMiddleware(asgi_app)

    async def run() -> None:
        scope = {"type": "websocket"}

        async def dummy_recv():
            return {}

        async def dummy_send(message):
            return None

        await mw(scope, dummy_recv, dummy_send)

    anyio.run(run)
    assert called["ok"] is True


def test_finished_receiver_exception_and_app_exception() -> None:
    s = Settings()
    app = create_app(s)

    # Receiver that raises on finished to exercise internal except
    def r_finished(sender, **kw):  # type: ignore[no-redef]
        raise RuntimeError("receiver failure")

    # Receiver that raises on got_request_exception
    def r_error(sender, **kw):  # type: ignore[no-redef]
        raise RuntimeError("error receiver failure")

    request_finished.connect(r_finished)
    got_request_exception.connect(r_error)

    @app.get("/ok2")
    def ok2() -> dict[str, str]:
        return {"x": "1"}

    @app.get("/bad")
    def bad() -> dict[str, str]:
        raise RuntimeError("boom")

    transport = httpx.ASGITransport(app=app)

    async def run() -> None:
        async with httpx.AsyncClient(transport=transport, base_url="http://t") as client:
            resp = await client.get("/ok2")
            assert resp.status_code == httpx.codes.OK
            with anyio.move_on_after(0):
                with suppress(Exception):
                    await client.get("/bad")

    try:
        anyio.run(run)
    finally:
        # cleanup to avoid leaking receivers across tests
        request_finished.disconnect(r_finished)
        got_request_exception.disconnect(r_error)
